@@ -3,6 +3,7 @@
  *  About    : Format strings
  *
  *  Copyright (C) 2006 YAMAMOTO Kengo <yamaken AT bp.iij4u.or.jp>
+ *  Copyright (c) 2007 SigScheme Project <uim AT freedesktop.org>
  *
  *  All rights reserved.
  *
@@ -69,12 +70,14 @@
 #include "sigschemeinternal.h"
 #if SCM_USE_MULTIBYTE_CHAR
 #include "encoding.h"
+#else
+#include "encoding-dummy.h"
 #endif
 
 /*=======================================
   File Local Macro Definitions
 =======================================*/
-#define PRETTY_PRINT_PROCEDURE_NAME "pretty-print"
+#define ERRMSG_INVALID_ESCSEQ "invalid escape sequence"
 
 #define MSG_SRFI48_DIRECTIVE_HELP                                             \
 "(format [<port>] <format-string> [<arg>...])\n"                              \
@@ -124,7 +127,7 @@
 #endif /* SCM_USE_SSCM_FORMAT_EXTENSION */
 
 #define NEWLINE_CHAR                                                         \
-    (SCM_NEWLINE_STR[sizeof(SCM_NEWLINE_STR) - sizeof("") - 1])
+    (SCM_NEWLINE_STR[sizeof(SCM_NEWLINE_STR) - 1 - sizeof("")])
 
 /*=======================================
   File Local Type Definitions
@@ -157,33 +160,25 @@ typedef const char *format_string_t;
 #define FORMAT_STR_PEEK(fmt)      (*(fmt))
 #endif /* SCM_USE_MULTIBYTE_CHAR */
 
-#define FORMAT_STR_SKIP_CHAR(fmt) (FORMAT_STR_READ(fmt), 0)
+#define FORMAT_STR_SKIP_CHAR(fmt) ((void)FORMAT_STR_READ(fmt))
+
 
 enum format_args_type {
     ARG_VA_LIST,
     ARG_SCM_LIST
 };
 
-#if 0
-/* Since this definition requires C99 va_copy() to work properly, the
- * alternative implementation is used to maximize portability. */
 struct format_args {
-    enum format_arg_type type;
+    enum format_args_type type;
     union {
-        va_list va;
-        ScmObj scm;
+        va_list *va;
+        ScmObj *scm;
     } lst;
 };
-#else
-union format_args {
-    va_list va;
-    ScmObj scm;
-};
-#endif
 
-#define POP_FORMAT_ARG(type, args)                                           \
-    (((type) == ARG_VA_LIST) ? va_arg((args)->va, ScmObj)                    \
-                             : MUST_POP_ARG((args)->scm))
+#define POP_FORMAT_ARG(args)                                                 \
+    (((args).type == ARG_VA_LIST) ? va_arg(*(args).lst.va, ScmObj)           \
+                                  : MUST_POP_ARG(*(args).lst.scm))
 
 /*=======================================
   Variable Definitions
@@ -216,13 +211,11 @@ static scm_ichar_t format_raw_c_directive(ScmObj port,
 static scm_ichar_t format_directive(ScmObj port, scm_ichar_t last_ch,
                                     enum ScmFormatCapability fcap,
                                     format_string_t *fmt,
-                                    enum format_args_type args_type,
-                                    union format_args *args);
+                                    struct format_args args);
 #endif
 static ScmObj format_internal(ScmObj port, enum ScmFormatCapability fcap,
                               const char *fmt,
-                              enum format_args_type args_type,
-                              union format_args *args);
+                              struct format_args args);
 
 /*=======================================
   Function Definitions
@@ -232,8 +225,7 @@ scm_init_format(void)
 {
     SCM_GLOBAL_VARS_INIT(static_format);
 
-    scm_gc_protect_with_init(&l_sym_pretty_print,
-                             scm_intern(PRETTY_PRINT_PROCEDURE_NAME));
+    scm_gc_protect_with_init(&l_sym_pretty_print, scm_intern("pretty-print"));
 }
 
 #if SCM_USE_MULTIBYTE_CHAR
@@ -245,23 +237,6 @@ format_str_peek(ScmMultibyteString mbs_fmt, const char *caller)
 }
 #endif /* SCM_USE_MULTIBYTE_CHAR */
 
-SCM_EXPORT void
-scm_pretty_print(ScmObj port, ScmObj obj)
-{
-    ScmObj proc_pretty_print;
-    DECLARE_INTERNAL_FUNCTION("scm_pretty_print");
-
-    /* FIXME: search pretty-print in current env */
-    proc_pretty_print = SCM_SYMBOL_VCELL(l_sym_pretty_print);
-
-    if (!EQ(proc_pretty_print, SCM_UNBOUND)) {
-        ENSURE_PROCEDURE(proc_pretty_print);
-        scm_call(proc_pretty_print, LIST_1(obj));
-    } else {
-        scm_write(port, obj);
-    }
-}
-
 static signed char
 read_width(format_string_t *fmt)
 {
@@ -290,6 +265,7 @@ read_width(format_string_t *fmt)
     return ret;
 }
 
+/* ([0-9]+(,[0-9]+)?)? */
 static ScmValueFormat
 read_number_prefix(enum ScmFormatCapability fcap, format_string_t *fmt)
 {
@@ -303,13 +279,19 @@ read_number_prefix(enum ScmFormatCapability fcap, format_string_t *fmt)
     if (c == '0' && (fcap & SCM_FMT_LEADING_ZEROS)) {
         FORMAT_STR_SKIP_CHAR(*fmt);
         vfmt.pad = '0';
+        vfmt.width = 0;
     }
     vfmt.width = read_width(fmt);
     c = FORMAT_STR_PEEK(*fmt);
 
     if (c == ',') {
+        if (vfmt.width < 0)
+            ERR(ERRMSG_INVALID_ESCSEQ ": ~~,");
         FORMAT_STR_SKIP_CHAR(*fmt);
         vfmt.frac_width = read_width(fmt);
+        if (vfmt.frac_width < 0)
+            ERR(ERRMSG_INVALID_ESCSEQ ": ~~~D,~C",
+                (int)vfmt.width, FORMAT_STR_PEEK(*fmt));
     }
 
     return vfmt;
@@ -327,7 +309,7 @@ format_int(ScmObj port, ScmValueFormat vfmt, uintmax_t n, int radix)
 
 #if SCM_USE_RAW_C_FORMAT
 /* returns '\0' if no valid directive handled */
-/* ([CP]|(0?[0-9]+(,0?[0-9]+)?)?(S|[MWQLGJTZ]?[UDXOB])) */
+/* ([CP]|([0-9]+(,[0-9]+)?)?(S|[MWQLGJTZ]?[UDXOB])) */
 static scm_ichar_t
 format_raw_c_directive(ScmObj port, format_string_t *fmt, va_list *args)
 {
@@ -447,7 +429,7 @@ format_raw_c_directive(ScmObj port, format_string_t *fmt, va_list *args)
     case 'U': /* Unsigned decimal */
         vfmt.signedp = scm_false;
         /* FALLTHROUGH */
-    case 'D': /* Decimal */
+    case 'D': /* signed Decimal */
         radix = 10;
         break;
 
@@ -467,7 +449,7 @@ format_raw_c_directive(ScmObj port, format_string_t *fmt, va_list *args)
         break;
 
     default:
-        /* no internal directives found */
+        /* no raw C directives found */
         if (FORMAT_STR_POS(*fmt) != orig_pos)
             *fmt = orig_fmt;
         return '\0';
@@ -487,14 +469,13 @@ format_raw_c_directive(ScmObj port, format_string_t *fmt, va_list *args)
 static scm_ichar_t
 format_directive(ScmObj port, scm_ichar_t last_ch,
                  enum ScmFormatCapability fcap,
-                 format_string_t *fmt,
-                 enum format_args_type args_type, union format_args *args)
+                 format_string_t *fmt, struct format_args args)
 {
     const void *orig_pos;
     char directive;
     scm_bool eolp;
 #if SCM_USE_SRFI48
-    ScmObj obj, indirect_fmt, indirect_args;
+    ScmObj obj, indirect_fmt, indirect_args, proc_pretty_print;
     scm_bool prefixedp;
     int radix;
     scm_int_t i;
@@ -515,18 +496,22 @@ format_directive(ScmObj port, scm_ichar_t last_ch,
         radix = -1;
         switch (directive) {
         case 'f': /* Fixed */
-            obj = POP_FORMAT_ARG(args_type, args);
+            obj = POP_FORMAT_ARG(args);
             if (STRINGP(obj)) {
                 for (i = SCM_STRING_LEN(obj); i < vfmt.width; i++)
                     scm_port_put_char(port, ' ');  /* ignore leading zero */
                 scm_display(port, obj);
-            } else {
-                if (!INTP(obj))
-                    ERR_OBJ("integer or string required but got", obj);
-                format_int(port, vfmt, SCM_INT_VALUE(obj), 10);
+                goto fin;
             }
-            goto fin;
+#if SCM_USE_INT
+            if (INTP(obj)) {
+                format_int(port, vfmt, SCM_INT_VALUE(obj), 10);
+                goto fin;
+            }
+#endif /* SCM_USE_INT */
+            ERR_OBJ("integer or string required but got", obj);
 
+#if SCM_USE_INT
         case 'd': /* Decimal */
             radix = 10;
             break;
@@ -542,14 +527,19 @@ format_directive(ScmObj port, scm_ichar_t last_ch,
         case 'b': /* Binary */
             radix = 2;
             break;
+#endif /* SCM_USE_INT */
 
         default:
             break;
         }
         if (radix > 0 && (!prefixedp || (fcap & SCM_FMT_PREFIXED_RADIX))) {
-            obj = POP_FORMAT_ARG(args_type, args);
+            obj = POP_FORMAT_ARG(args);
+#if SCM_USE_INT
             ENSURE_INT(obj);
             format_int(port, vfmt, SCM_INT_VALUE(obj), radix);
+#else /* SCM_USE_INT */
+            ERR("integer feature is not enabled");
+#endif /* SCM_USE_INT */
             goto fin;
         }
     }
@@ -561,11 +551,11 @@ format_directive(ScmObj port, scm_ichar_t last_ch,
     if (fcap & SCM_FMT_SRFI28) {
         switch (directive) {
         case 'a': /* Any */
-            scm_display(port, POP_FORMAT_ARG(args_type, args));
+            scm_display(port, POP_FORMAT_ARG(args));
             goto fin;
 
         case 's': /* Slashified */
-            scm_write(port, POP_FORMAT_ARG(args_type, args));
+            scm_write(port, POP_FORMAT_ARG(args));
             goto fin;
 
         case '%': /* Newline */
@@ -585,30 +575,42 @@ format_directive(ScmObj port, scm_ichar_t last_ch,
 #if SCM_USE_SRFI48
     if (fcap & SCM_FMT_SRFI48_ADDENDUM) {
         switch (directive) {
+#if SCM_USE_SRFI38
         case 'w': /* WriteCircular */
-            scm_write_ss(port, POP_FORMAT_ARG(args_type, args));
+            scm_write_ss(port, POP_FORMAT_ARG(args));
             goto fin;
+#endif /* SCM_USE_SRFI38 */
 
         case 'y': /* Yuppify */
-            scm_pretty_print(port, POP_FORMAT_ARG(args_type, args));
+            proc_pretty_print = SCM_SYMBOL_VCELL(l_sym_pretty_print);
+            if (EQ(proc_pretty_print, SCM_UNBOUND)) {
+                /* called internally when (use srfi-48) is not evaluated yet */
+                scm_write(port, POP_FORMAT_ARG(args));
+            } else {
+                ENSURE_PROCEDURE(proc_pretty_print);
+                obj = POP_FORMAT_ARG(args);
+                scm_call(proc_pretty_print, LIST_2(obj, port));
+            }
             goto fin;
 
         case 'k': /* Indirection (backward compatability) */
         case '?': /* Indirection */
-            indirect_fmt = POP_FORMAT_ARG(args_type, args);
+            indirect_fmt = POP_FORMAT_ARG(args);
             ENSURE_STRING(indirect_fmt);
-            indirect_args = POP_FORMAT_ARG(args_type, args);
+            indirect_args = POP_FORMAT_ARG(args);
             ENSURE_LIST(indirect_args);
             scm_lformat(port,
                         fcap & ~SCM_FMT_RAW_C,
                         SCM_STRING_STR(indirect_fmt), indirect_args);
             goto fin;
 
+#if SCM_USE_CHAR
         case 'c': /* Character */
-            obj = POP_FORMAT_ARG(args_type, args);
+            obj = POP_FORMAT_ARG(args);
             ENSURE_CHAR(obj);
             scm_port_put_char(port, SCM_CHAR_VALUE(obj));
             goto fin;
+#endif /* SCM_USE_CHAR */
 
         case 't': /* Tab */
             scm_port_put_char(port, '\t');
@@ -641,7 +643,7 @@ format_directive(ScmObj port, scm_ichar_t last_ch,
 
     /* Although SRFI-48 does not specified about unknown directives, the
      * reference implementation treats it as error. */
-    ERR("invalid escape sequence: ~~~C", (scm_ichar_t)directive);
+    ERR(ERRMSG_INVALID_ESCSEQ ": ~~~C", (scm_ichar_t)directive);
 
  fin:
     FORMAT_STR_SKIP_CHAR(*fmt);
@@ -651,8 +653,7 @@ format_directive(ScmObj port, scm_ichar_t last_ch,
 
 static ScmObj
 format_internal(ScmObj port, enum ScmFormatCapability fcap,
-                const char *fmt,
-                enum format_args_type args_type, union format_args *args)
+                const char *fmt, struct format_args args)
 {
     scm_ichar_t c, last_c, handled;
     format_string_t cur;
@@ -660,8 +661,12 @@ format_internal(ScmObj port, enum ScmFormatCapability fcap,
     DECLARE_INTERNAL_FUNCTION("format");
 
     if (FALSEP(port)) {
+#if SCM_USE_SRFI6
         port = scm_p_srfi6_open_output_string();
         implicit_portp = scm_true;
+#else
+        ERR("format to string needs SRFI-6 feature");
+#endif
     } else if (EQ(port, SCM_TRUE)) {
         port = scm_out;
         implicit_portp = scm_false;
@@ -678,8 +683,8 @@ format_internal(ScmObj port, enum ScmFormatCapability fcap,
         if (c == '~') {
 #if SCM_USE_RAW_C_FORMAT
             if (fcap & SCM_FMT_RAW_C) {
-                SCM_ASSERT(args_type == ARG_VA_LIST);
-                handled = format_raw_c_directive(port, &cur, &args->va);
+                SCM_ASSERT(args.type == ARG_VA_LIST);
+                handled = format_raw_c_directive(port, &cur, args.lst.va);
                 if (handled) {
                     last_c = handled;
                     continue;
@@ -689,40 +694,70 @@ format_internal(ScmObj port, enum ScmFormatCapability fcap,
 #endif /* SCM_USE_RAW_C_FORMAT */
 #if SCM_USE_SRFI28
             if (fcap & (SCM_FMT_SRFI28 | SCM_FMT_SRFI48 | SCM_FMT_SSCM)) {
-                SCM_ASSERT(args_type == ARG_VA_LIST
-                           || args_type == ARG_SCM_LIST);
-                last_c = format_directive(port, last_c, fcap, &cur,
-                                          args_type, args);
+                SCM_ASSERT(args.type == ARG_VA_LIST
+                           || args.type == ARG_SCM_LIST);
+                last_c = format_directive(port, last_c, fcap, &cur, args);
                 continue;
             }
 #endif /* SCM_USE_SRFI28 */
-            SCM_ASSERT(scm_false);
+            SCM_NOTREACHED;
         } else {
             scm_port_put_char(port, c);
             last_c = c;
         }
     }
 
-    if (args_type == ARG_SCM_LIST)
-        ENSURE_NO_MORE_ARG(args->scm);
+    if (args.type == ARG_SCM_LIST)
+        ENSURE_NO_MORE_ARG(*args.lst.scm);
+#if SCM_USE_SRFI6
     return (implicit_portp) ? scm_p_srfi6_get_output_string(port) : SCM_UNDEF;
+#else
+    return SCM_UNDEF;
+#endif
 }
 
 SCM_EXPORT ScmObj
 scm_lformat(ScmObj port,
             enum ScmFormatCapability fcap, const char *fmt, ScmObj scm_args)
 {
-    return format_internal(port, fcap, fmt,
-                           ARG_SCM_LIST, (union format_args *)&scm_args);
+    struct format_args args;
+
+    args.type = ARG_SCM_LIST;
+    args.lst.scm = &scm_args;
+    return format_internal(port, fcap, fmt, args);
 }
 
 SCM_EXPORT ScmObj
 scm_vformat(ScmObj port,
             enum ScmFormatCapability fcap, const char *fmt, va_list c_args)
 {
-    return format_internal(port, fcap, fmt,
-                           ARG_VA_LIST,
-                           (union format_args *)(uintptr_t)&c_args);
+    struct format_args args;
+
+    args.type = ARG_VA_LIST;
+#if HAVE_REFERENCEABLE_PASSED_VA_LIST
+    /* { va_list ap; return &ap; } and f(va_list ap) { return &ap; } returns
+     * same value */
+    args.lst.va = &c_args;
+#elif HAVE_AUTOREFERRED_PASSED_VA_LIST
+    /* f(va_list ap) { return &ap; } returns invalid value */
+    /*
+     * x86_64 on gcc and some environemnts behaves such a way. See following
+     * bug reports of gcc for further information.
+     *
+     * http://gcc.gnu.org/bugzilla/show_bug.cgi?id=14557
+     * http://gcc.gnu.org/bugzilla/show_bug.cgi?id=20951
+     *
+     * To avoid taking an address of va_list that passed as a function argument
+     * is the best way to maximize the portability. But since it requires a
+     * combined function of format_internal(), format_raw_c_directive() and
+     * format_directive() which considerably makes the maintainability of
+     * format.c lost, I adopted this hack.  -- YamaKen 2006-12-07
+     */
+    args.lst.va = (va_list *)c_args;
+#else
+#error "This platform is not supported"
+#endif
+    return format_internal(port, fcap, fmt, args);
 }
 
 SCM_EXPORT ScmObj
