@@ -3,6 +3,7 @@
  *  About    : SRFI-34 Exception Handling for Programs
  *
  *  Copyright (C) 2005-2006 YAMAMOTO Kengo <yamaken AT bp.iij4u.or.jp>
+ *  Copyright (c) 2007 SigScheme Project <uim AT freedesktop.org>
  *
  *  All rights reserved.
  *
@@ -36,7 +37,7 @@
  * This file implements C-version of the reference implementation written in
  * the SRFI-34 specification. All parts are written in C since:
  *
- * - SigScheme doesn't have a hygienic-macros feature (yet)
+ * - SigScheme's hygienic-macros feature is not efficient yet
  *
  * - To avoid namespace pollution (with-exception-handlers, guard-aux, etc),
  *   since SigScheme doesn't have a module or namespace feature (yet)
@@ -52,7 +53,6 @@
 =======================================*/
 #define USE_WITH_SIGSCHEME_FATAL_ERROR 1
 
-#define ERRMSG_UNHANDLED_EXCEPTION "unhandled exception"
 #define ERRMSG_HANDLER_RETURNED    "handler returned"
 #define ERRMSG_FALLBACK_EXHAUSTED  "fallback handler exhausted"
 
@@ -131,6 +131,7 @@ SCM_DEFINE_STATIC_VARS(static_srfi34);
   File Local Function Declarations
 =======================================*/
 static ScmObj raw_quote(ScmObj datum, ScmObj env);
+static ScmObj enclose(ScmObj exp, ScmObj env);
 static ScmObj set_cur_handlers(ScmObj handlers, ScmObj env);
 static ScmObj with_exception_handlers(ScmObj new_handlers, ScmObj thunk);
 static ScmObj guard_internal(ScmObj q_guard_k, ScmObj env);
@@ -157,7 +158,7 @@ scm_initialize_srfi34(void)
          var < &vars[N_GLOBAL_SCMOBJ];
          var++)
     {
-        scm_gc_protect_with_init(var, SCM_FALSE);
+        scm_gc_protect_with_init(var, SCM_UNDEF);
     }
 
     l_errmsg_unhandled_exception = CONST_STRING(ERRMSG_UNHANDLED_EXCEPTION);
@@ -175,9 +176,12 @@ scm_initialize_srfi34(void)
     l_sym_handler_k  = scm_intern("handler-k");
 
     /* prepare procedures and syntaxes */
-    l_syn_apply   = scm_symbol_value(scm_intern("apply"),  SCM_INTERACTION_ENV);
-    l_proc_values = scm_symbol_value(scm_intern("values"), SCM_INTERACTION_ENV);
+    l_syn_apply
+        = scm_symbol_value(scm_intern("apply"),  SCM_INTERACTION_ENV);
+    l_proc_values
+        = scm_symbol_value(scm_intern("values"), SCM_INTERACTION_ENV);
 
+    SCM_ASSERT_FUNCTYPE(scm_syntax_fixed_1,         &raw_quote);
     SCM_ASSERT_FUNCTYPE(scm_syntax_fixed_1,         &set_cur_handlers);
     SCM_ASSERT_FUNCTYPE(scm_procedure_fixed_2,      &with_exception_handlers);
     SCM_ASSERT_FUNCTYPE(scm_syntax_fixed_1,         &guard_internal);
@@ -200,6 +204,11 @@ scm_initialize_srfi34(void)
     l_syn_guard_body
         = MAKE_FUNC(SCM_SYNTAX_FIXED_TAILREC_0, &guard_body);
 
+    /*
+     * The 'error' procedure should not be invoked directly by
+     * scm_p_srfi23_error(), to allow dynamic redifinition, and keep SRFI-23
+     * implementation abstract.
+     */
 #if USE_WITH_SIGSCHEME_FATAL_ERROR
     l_proc_fallback_handler
         = scm_s_lambda(LIST_1(l_sym_condition),
@@ -213,11 +222,6 @@ scm_initialize_srfi34(void)
                                             l_sym_condition))),
                        SCM_INTERACTION_ENV);
 #else /* USE_WITH_SIGSCHEME_FATAL_ERROR */
-    /*
-     * The 'error' procedure should not be invoked directly by
-     * scm_p_srfi23_error(), to allow dynamic redifinition, and keep SRFI-23
-     * implementation abstract.
-     */
     l_proc_fallback_handler
         = scm_s_lambda(LIST_1(l_sym_condition),
                        LIST_1(LIST_3(l_sym_error,
@@ -226,7 +230,7 @@ scm_initialize_srfi34(void)
                        SCM_INTERACTION_ENV);
 #endif /* USE_WITH_SIGSCHEME_FATAL_ERROR */
 
-    scm_register_funcs(scm_srfi34_func_info_table);
+    scm_register_funcs(scm_functable_srfi34);
 
     l_current_exception_handlers = LIST_1(l_proc_fallback_handler);
 }
@@ -240,6 +244,12 @@ raw_quote(ScmObj datum, ScmObj env)
     return datum;
 }
 
+static ScmObj
+enclose(ScmObj exp, ScmObj env)
+{
+    return scm_s_lambda(SCM_NULL, LIST_1(exp), env);
+}
+
 static ScmObj
 set_cur_handlers(ScmObj handlers, ScmObj env)
 {
@@ -303,8 +313,9 @@ scm_p_srfi34_raise(ScmObj obj)
     obj = LIST_2(SYM_QUOTE, obj);
     thunk = scm_s_lambda(SCM_NULL,
                          LIST_2(LIST_2(handler, obj),
-                                LIST_3(l_sym_error,
-                                       l_errmsg_handler_returned, obj)),
+                                LIST_4(l_sym_error,
+                                       l_errmsg_handler_returned,
+                                       handler, obj)),
                          SCM_INTERACTION_ENV);
     return with_exception_handlers(rest_handlers, thunk);
 }
@@ -366,7 +377,6 @@ guard_handler(ScmObj q_condition, ScmEvalState *eval_state)
     return scm_call(ret, SCM_NULL);
 }
 
-/* assumes that scm_s_delay() returns a closure */
 static ScmObj
 delay(ScmObj evaled_obj, ScmObj env)
 {
@@ -374,16 +384,15 @@ delay(ScmObj evaled_obj, ScmObj env)
 
     if (VALUEPACKETP(evaled_obj)) {
         vals = SCM_VALUEPACKET_VALUES(evaled_obj);
-        return scm_s_delay(LIST_3(l_syn_apply,
-                                  l_proc_values,
-                                  LIST_2(l_syn_raw_quote, vals)),
-                           env);
+        return enclose(LIST_3(l_syn_apply,
+                              l_proc_values,
+                              LIST_2(l_syn_raw_quote, vals)),
+                       env);
     } else {
-        return scm_s_delay(LIST_2(l_syn_raw_quote, evaled_obj), env);
+        return enclose(LIST_2(l_syn_raw_quote, evaled_obj), env);
     }
 }
 
-/* assumes that scm_s_delay() returns a closure */
 static ScmObj
 guard_handler_body(ScmObj q_handler_k, ScmObj env)
 {
@@ -402,19 +411,18 @@ guard_handler_body(ScmObj q_handler_k, ScmObj env)
     sym_var = CAR(cond_catch);
     clauses = CDR(cond_catch);
     ENSURE_SYMBOL(sym_var);
-    cond_env = scm_extend_environment(LIST_1(sym_var),
-                                      LIST_1(condition),
-                                      lex_env);
+    cond_env
+        = scm_extend_environment(LIST_1(sym_var), LIST_1(condition), lex_env);
     SCM_EVAL_STATE_INIT1(eval_state, cond_env);
-    caught = scm_s_cond_internal(clauses, SCM_INVALID, &eval_state);
+    caught = scm_s_cond_internal(clauses, &eval_state);
 
     if (VALIDP(caught)) {
         if (eval_state.ret_type == SCM_VALTYPE_NEED_EVAL)
             caught = EVAL(caught, cond_env);
         scm_call_continuation(guard_k, delay(caught, cond_env));
     } else {
-        reraise = scm_s_delay(LIST_2(l_sym_raise, LIST_2(SYM_QUOTE, condition)),
-                              cond_env);
+        reraise = enclose(LIST_2(l_sym_raise, LIST_2(SYM_QUOTE, condition)),
+                          cond_env);
         scm_call_continuation(handler_k, reraise);
     }
     /* NOTREACHED */
