@@ -3,7 +3,10 @@
  *  About    : S-Expression reader
  *
  *  Copyright (C) 2000-2005 Shiro Kawai <shiro AT acm.org>
- *  Copyright (C) 2005-2006 Kazuki Ohta <mover AT hct.zaq.ne.jp>
+ *  Copyright (C) 2005      Kazuki Ohta <mover AT hct.zaq.ne.jp>
+ *  Copyright (C) 2005-2006 Jun Inoue <jun.lambda AT gmail.com>
+ *  Copyright (C) 2005-2006 YAMAMOTO Kengo <yamaken AT bp.iij4u.or.jp>
+ *  Copyright (c) 2007 SigScheme Project <uim AT freedesktop.org>
  *
  *  All rights reserved.
  *
@@ -33,6 +36,17 @@
  *  ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ===========================================================================*/
 
+/*
+ * FIXME: Replace obsoleted SRFI-75 with latest R6RS specifications once it has
+ * been stabilized.  -- YamaKen 2006-11-28
+ *
+ * - Remove #\uXXXX and #\UXXXXXXXX literals
+ * - Support variable-length #\xXX literal
+ * - Support character category validation for identifiers
+ * - Disable #\newline on R6RS-compatible mode
+ * - Confirm symbol escape syntax (not defined in R6RS yet)
+ */
+
 /*
  * ChangeLog
  *
@@ -125,6 +139,33 @@
  * <digit 16> --> <digit 10> | a | b | c | d | e | f
  */
 
+/*
+ * Although R5RS defined number literals as above, SigScheme only supports
+ * these truncated forms. See "R5RS conformance: Numbers: Literals" section of
+ * doc/spec.txt.
+ *
+ * <number> --> <num 2>| <num 8>
+ *      | <num 10>| <num 16>
+ * 
+ * <num R> --> <prefix R> <complex R>
+ * <complex R> --> <real R>
+ * <real R> --> <sign> <ureal R>
+ * <ureal R> --> <uinteger R>
+ * <uinteger R> --> <digit R>+ #*   ;; '#' must not occur
+ * <prefix R> --> <radix R>
+ *
+ * <sign> --> <empty>  | + |  -
+ * <radix 2> --> #b
+ * <radix 8> --> #o
+ * <radix 10> --> <empty> | #d
+ * <radix 16> --> #x
+ * <digit 2> --> 0 | 1
+ * <digit 8> --> 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7
+ * <digit 10> --> <digit>
+ * <digit 16> --> <digit 10> | a | b | c | d | e | f 
+ * <digit> --> 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9
+ */
+
 #include <config.h>
 
 #include <limits.h>
@@ -143,17 +184,15 @@
 #define OK 0
 #define TOKEN_BUF_EXCEEDED (-1)
 
-#define MB_MAX_SIZE (SCM_MB_MAX_LEN + sizeof(""))
-
 /* can accept "backspace" of R5RS and "U0010FFFF" of SRFI-75 */
 #define CHAR_LITERAL_LEN_MAX (sizeof("backspace") - sizeof(""))
 
 /* #b-010101... */
-#define INT_LITERAL_LEN_MAX (sizeof((char)'-') + SCM_INT_BITS)
+#define INT_LITERAL_LEN_MAX (sizeof("-0") + SCM_INT_BITS - sizeof(""))
 
 #define DISCARD_LOOKAHEAD(port) (scm_port_get_char(port))
 
-/* accepts EOF */
+/* accepts SCM_ICHAR_EOF */
 #define ICHAR_ASCII_CLASS(c)                                                 \
     (ICHAR_ASCIIP(c) ? scm_char_class_table[c] : SCM_CH_INVALID)
 #define ICHAR_CLASS(c)                                                       \
@@ -335,24 +374,30 @@ static const unsigned char scm_char_class_table[] = {
   File Local Function Declarations
 =======================================*/
 static scm_ichar_t skip_comment_and_space(ScmObj port);
-static void   read_sequence(ScmObj port, char *buf, int len);
 static size_t read_token(ScmObj port, int *err,
                          char *buf, size_t buf_size, enum ScmCharClass delim);
 
 static ScmObj read_sexpression(ScmObj port);
-static ScmObj read_list(ScmObj port, scm_ichar_t closeParen);
+static ScmObj read_list(ScmObj port, scm_ichar_t closing_paren);
 #if SCM_USE_SRFI75
+static void        read_sequence(ScmObj port, char *buf, int len);
 static scm_ichar_t parse_unicode_sequence(const char *seq, int len);
 static scm_ichar_t read_unicode_sequence(ScmObj port, char prefix);
-#endif
+#endif /* SCM_USE_SRFI75 */
+#if SCM_USE_CHAR
 static ScmObj read_char(ScmObj port);
+#endif /* SCM_USE_CHAR */
+#if SCM_USE_STRING
 static ScmObj read_string(ScmObj port);
+#endif /* SCM_USE_STRING */
 static ScmObj read_symbol(ScmObj port);
-static ScmObj read_number_or_symbol(ScmObj port);
+static ScmObj read_number_or_peculiar(ScmObj port);
+#if SCM_USE_NUMBER
 static ScmObj parse_number(ScmObj port,
                            char *buf, size_t buf_size, char prefix);
 static ScmObj read_number(ScmObj port, char prefix);
-static ScmObj read_quote(ScmObj port, ScmObj quoter);
+#endif /* SCM_USE_NUMBER */
+static ScmObj read_quoted(ScmObj port, ScmObj quoter);
 
 /*=======================================
   Function Definitions
@@ -377,17 +422,6 @@ scm_read(ScmObj port)
     return sexp;
 }
 
-SCM_EXPORT ScmObj
-scm_read_char(ScmObj port)
-{
-    DECLARE_INTERNAL_FUNCTION("scm_read_char");
-
-    ENSURE_PORT(port);
-
-    return read_char(port);
-}
-
-
 static scm_ichar_t
 skip_comment_and_space(ScmObj port)
 {
@@ -400,14 +434,14 @@ skip_comment_and_space(ScmObj port)
         case LEX_ST_NORMAL:
             if (c == ';')
                 state = LEX_ST_COMMENT;
-            else if (!ICHAR_WHITESPACEP(c) || c == EOF)
+            else if (!ICHAR_WHITESPACEP(c) || c == SCM_ICHAR_EOF)
                 return c;  /* peeked */
             break;
 
         case LEX_ST_COMMENT:
             if (c == '\n' || c == '\r')
                 state = LEX_ST_NORMAL;
-            else if (c == EOF)
+            else if (c == SCM_ICHAR_EOF)
                 return c;  /* peeked */
             break;
         }
@@ -415,29 +449,13 @@ skip_comment_and_space(ScmObj port)
     }
 }
 
-static void
-read_sequence(ScmObj port, char *buf, int len)
-{
-    scm_ichar_t c;
-    char *p;
-    DECLARE_INTERNAL_FUNCTION("read");
-
-    for (p = buf; p < &buf[len]; p++) {
-        c = scm_port_get_char(port);
-        if (c == EOF)
-            ERR("unexpected EOF");
-        if (!ICHAR_ASCIIP(c))
-            ERR("unexpected non-ASCII char");
-        *p = c;
-    }
-    buf[len] = '\0';
-}
-
 static size_t
 read_token(ScmObj port, int *err,
            char *buf, size_t buf_size, enum ScmCharClass delim)
 {
+#if SCM_USE_SRFI75
     ScmCharCodec *codec;
+#endif
     enum ScmCharClass ch_class;
     scm_ichar_t c;
     size_t len;
@@ -450,10 +468,10 @@ read_token(ScmObj port, int *err,
         CDBG((SCM_DBG_PARSER, "c = ~C", c));
 
         if (p == buf) {
-            if (c == EOF)
+            if (c == SCM_ICHAR_EOF)
                 ERR("unexpected EOF at a token");
         } else {
-            if (ch_class & delim || c == EOF) {
+            if (ch_class & delim || c == SCM_ICHAR_EOF) {
                 *err = OK;
                 break;
             }
@@ -493,7 +511,9 @@ read_token(ScmObj port, int *err,
 static ScmObj
 read_sexpression(ScmObj port)
 {
+#if SCM_USE_VECTOR
     ScmObj ret;
+#endif
     enum ScmCharClass ch_class;
     scm_ichar_t c;
     DECLARE_INTERNAL_FUNCTION("read");
@@ -510,19 +530,23 @@ read_sexpression(ScmObj port)
             return read_symbol(port);
 
         if (ch_class & (SCM_CH_DIGIT | SCM_CH_PECULIAR_IDENTIFIER_CAND))
-            return read_number_or_symbol(port);
+            return read_number_or_peculiar(port);
 
         /* case labels are ordered by appearance rate and penalty cost */
+        SCM_ASSERT(ch_class == SCM_CH_TOKEN_INITIAL || c == SCM_ICHAR_EOF);
+        SCM_ASSERT(c != ';');
         DISCARD_LOOKAHEAD(port);
         switch (c) {
         case '(':
             return read_list(port, ')');
 
+#if SCM_USE_STRING
         case '\"':
             return read_string(port);
+#endif
 
         case '\'':
-            return read_quote(port, SYM_QUOTE);
+            return read_quoted(port, SYM_QUOTE);
 
         case '#':
             c = scm_port_get_char(port);
@@ -531,69 +555,92 @@ read_sexpression(ScmObj port)
                 return SCM_TRUE;
             case 'f':
                 return SCM_FALSE;
+#if SCM_USE_VECTOR
             case '(':
                 ret = scm_p_list2vector(read_list(port, ')'));
 #if SCM_CONST_VECTOR_LITERAL
                 SCM_VECTOR_SET_IMMUTABLE(ret);
 #endif
                 return ret;
+#endif /* SCM_USE_VECTOR */
+#if SCM_USE_CHAR
             case '\\':
                 return read_char(port);
+#endif
+#if SCM_USE_NUMBER
+            /* TODO: support exactness prefixes 'i' and 'e' */
             case 'b': case 'o': case 'd': case 'x':
                 return read_number(port, c);
-            case EOF:
+#endif
+            case SCM_ICHAR_EOF:
                 ERR("EOF in #");
+                /* NOTREACHED */
             default:
-                ERR("Unsupported # notation: ~C", c);
+                ERR("unsupported # notation: ~C", c);
+                /* NOTREACHED */
             }
-            break;
+            /* NOTREACHED */
 
         case '`':
-            return read_quote(port, SYM_QUASIQUOTE);
+            return read_quoted(port, SYM_QUASIQUOTE);
 
         case ',':
             c = scm_port_peek_char(port);
             switch (c) {
-            case EOF:
+            case SCM_ICHAR_EOF:
                 ERR("EOF in unquote");
                 /* NOTREACHED */
 
             case '@':
                 DISCARD_LOOKAHEAD(port);
-                return read_quote(port, SYM_UNQUOTE_SPLICING);
+                return read_quoted(port, SYM_UNQUOTE_SPLICING);
 
             default:
-                return read_quote(port, SYM_UNQUOTE);
+                return read_quoted(port, SYM_UNQUOTE);
             }
+            /* NOTREACHED */
 
         case ')':
             ERR("unexpected ')'");
             /* NOTREACHED */
 
-        case EOF:
+        case SCM_ICHAR_EOF:
             return SCM_EOF;
 
+        case '|':
+        case '[':
+        case ']':
+        case '{':
+        case '}':
+            ERR("reserved notation: ~C", c);
+            /* NOTREACHED */
+
         default:
-            SCM_ASSERT(scm_false);
+            SCM_NOTREACHED;
         }
     }
 }
 
 static ScmObj
-read_list(ScmObj port, scm_ichar_t closeParen)
+read_list(ScmObj port, scm_ichar_t closing_paren)
 {
     ScmObj lst, elm, cdr;
     ScmQueue q;
+#if SCM_DEBUG
     ScmBaseCharPort *basecport;
+    size_t start_line, cur_line;
+#endif
     scm_ichar_t c;
-    int err, start_line, cur_line;
+    int err;
     char dot_buf[sizeof("...")];
     DECLARE_INTERNAL_FUNCTION("read");
 
+#if SCM_DEBUG
     CDBG((SCM_DBG_PARSER, "read_list"));
     basecport = SCM_PORT_TRY_DYNAMIC_CAST(ScmBaseCharPort,
                                           SCM_PORT_IMPL(port));
-    start_line = (basecport) ? ScmBaseCharPort_line_number(basecport) : -1;
+    start_line = (basecport) ? ScmBaseCharPort_line_number(basecport) : 0;
+#endif
 
     for (lst = SCM_NULL, SCM_QUEUE_POINT_TO(q, lst);
          ;
@@ -608,22 +655,22 @@ read_list(ScmObj port, scm_ichar_t closeParen)
 
         CDBG((SCM_DBG_PARSER, "read_list c = [~C]", c));
 
-        if (c == EOF) {
-            if (basecport) {
+        if (c == SCM_ICHAR_EOF) {
+#if SCM_DEBUG
+            if (basecport && start_line) {
                 cur_line = ScmBaseCharPort_line_number(basecport);
-                ERR("EOF inside list at line ~D (starting from line ~D)",
+                ERR("EOF inside list at line ~ZU (started from line ~ZU)",
                     cur_line, start_line);
-            } else {
+            } else
+#endif
                 ERR("EOF inside list");
-            }
-        } else if (c == closeParen) {
+        } else if (c == closing_paren) {
             DISCARD_LOOKAHEAD(port);
             return lst;
         } else if (c == '.') {
             /* Since expressions that beginning with a dot are limited to '.',
-             * '...' and numbers in R5RS (See "7.1.1 Lexical structure"), fixed
-             * size buffer can safely buffer them.
-             */
+             * '...' and numbers in R5RS (See "7.1.1 Lexical structure"), the
+             * fixed size buffer can safely buffer them. */
             read_token(port, &err, dot_buf, sizeof(dot_buf), SCM_CH_DELIMITER);
 
             if (dot_buf[1] == '\0') {
@@ -633,8 +680,7 @@ read_list(ScmObj port, scm_ichar_t closeParen)
                  * (e.g. '("foo"."bar") is parsed as 3 element list which 2nd
                  * elem is dot as symbol). To avoid introducing such
                  * incompatibility problem into codes of SigScheme users,
-                 * require explicit whitespace around the dot.
-                 */
+                 * require explicit whitespace around the dot. */
                 c = scm_port_peek_char(port);
                 if (!ICHAR_WHITESPACEP(c))
                     ERR("implicit dot delimitation is disabled to avoid compatibility problem");
@@ -645,13 +691,13 @@ read_list(ScmObj port, scm_ichar_t closeParen)
                 cdr = read_sexpression(port);
                 c = skip_comment_and_space(port);
                 DISCARD_LOOKAHEAD(port);
-                if (c != ')')
+                if (c != closing_paren)
                     ERR("bad dot syntax");
 
                 SCM_QUEUE_SLOPPY_APPEND(q, cdr);
                 return lst;
             } else if (strcmp(dot_buf, "...") == 0) {
-                elm = scm_intern(dot_buf);
+                elm = SYM_ELLIPSIS;
             } else {
                 ERR("bad dot syntax");
             }
@@ -706,6 +752,24 @@ parse_unicode_sequence(const char *seq, int len)
     return c;
 }
 
+static void
+read_sequence(ScmObj port, char *buf, int len)
+{
+    scm_ichar_t c;
+    char *p;
+    DECLARE_INTERNAL_FUNCTION("read");
+
+    for (p = buf; p < &buf[len]; p++) {
+        c = scm_port_get_char(port);
+        if (c == SCM_ICHAR_EOF)
+            ERR("unexpected EOF");
+        if (!ICHAR_ASCIIP(c))
+            ERR("unexpected non-ASCII char");
+        *p = c;
+    }
+    buf[len] = '\0';
+}
+
 static scm_ichar_t
 read_unicode_sequence(ScmObj port, char prefix)
 {
@@ -717,7 +781,7 @@ read_unicode_sequence(ScmObj port, char prefix)
     case 'u': len = 4; break;
     case 'U': len = 8; break;
     default:
-        SCM_ASSERT(scm_false);
+        SCM_NOTREACHED;
     }
     seq[0] = prefix;
     read_sequence(port, &seq[1], len);
@@ -725,11 +789,14 @@ read_unicode_sequence(ScmObj port, char prefix)
 }
 #endif /* SCM_USE_SRFI75 */
 
+#if SCM_USE_CHAR
 static ScmObj
 read_char(ScmObj port)
 {
     const ScmSpecialCharInfo *info;
+#if SCM_USE_SRFI75
     ScmCharCodec *codec;
+#endif
     size_t len;
     scm_ichar_t c, next;
 #if SCM_USE_SRFI75
@@ -739,10 +806,10 @@ read_char(ScmObj port)
     char buf[CHAR_LITERAL_LEN_MAX + sizeof("")];
     DECLARE_INTERNAL_FUNCTION("read");
 
-    /* plain char (multibyte-ready) */
+    /* raw char (multibyte-ready) */
     c = scm_port_get_char(port);
     next = scm_port_peek_char(port);
-    if (ICHAR_ASCII_CLASS(next) & SCM_CH_DELIMITER || next == EOF)
+    if (ICHAR_ASCII_CLASS(next) & SCM_CH_DELIMITER || next == SCM_ICHAR_EOF)
         return MAKE_CHAR(c);
 #if SCM_USE_SRFI75
     else if (!ICHAR_ASCIIP(c))
@@ -776,7 +843,9 @@ read_char(ScmObj port)
     }
     ERR("invalid character literal: #\\~S", buf);
 }
+#endif /* SCM_USE_CHAR */
 
+#if SCM_USE_STRING
 static ScmObj
 read_string(ScmObj port)
 {
@@ -805,15 +874,14 @@ read_string(ScmObj port)
         CDBG((SCM_DBG_PARSER, "read_string c = ~C", c));
 
         switch (c) {
-        case EOF:
-            LBUF_EXTEND(lbuf, SCM_LBUF_F_STRING, offset + 1);
-            *p = '\0';
-            ERR("EOF in string: \"~S<eof>", LBUF_BUF(lbuf));
-            break;
+        case SCM_ICHAR_EOF:
+            LBUF_FREE(lbuf);
+            ERR("EOF in string");
+            /* NOTREACHED */
 
         case '\"':
             LBUF_EXTEND(lbuf, SCM_LBUF_F_STRING, offset + 1);
-            *p = '\0';
+            LBUF_BUF(lbuf)[offset] = '\0';
             obj = MAKE_IMMUTABLE_STRING_COPYING(LBUF_BUF(lbuf), len);
             LBUF_FREE(lbuf);
             return obj;
@@ -825,7 +893,8 @@ read_string(ScmObj port)
                 if (c != 'x' && SCM_CHARCODEC_CCS(codec) != SCM_CCS_UCS4)
                     ERR_OBJ("Unicode char sequence on non-Unicode port", port);
                 c = read_unicode_sequence(port, c);
-                LBUF_EXTEND(lbuf, SCM_LBUF_F_STRING, offset + MB_MAX_SIZE);
+                LBUF_EXTEND(lbuf, SCM_LBUF_F_STRING,
+                            offset + SCM_MB_CHAR_BUF_SIZE);
                 p = &LBUF_BUF(lbuf)[offset];
                 p = SCM_CHARCODEC_INT2STR(codec, p, c, SCM_MB_STATELESS);
                 if (!p)
@@ -850,12 +919,17 @@ read_string(ScmObj port)
             break;
 
         default:
-            LBUF_EXTEND(lbuf, SCM_LBUF_F_STRING, offset + MB_MAX_SIZE);
+            LBUF_EXTEND(lbuf, SCM_LBUF_F_STRING,
+                        offset + SCM_MB_CHAR_BUF_SIZE);
             p = &LBUF_BUF(lbuf)[offset];
+#if SCM_USE_SRFI75
             /* FIXME: support stateful encoding */
             p = SCM_CHARCODEC_INT2STR(codec, p, c, SCM_MB_STATELESS);
             if (!p)
                 ERR("invalid char in string: 0x~MX", (scm_int_t)c);
+#else
+            *p++ = c;
+#endif
             break;
         }
 #if !SCM_USE_NULL_CAPABLE_STRING
@@ -863,10 +937,13 @@ read_string(ScmObj port)
             ERR(SCM_ERRMSG_NULL_IN_STRING);
 #endif
     }
+#if 0
     LBUF_END(lbuf)[-1] = '\0';
     ERR("too long string: \"~S\"", LBUF_BUF(lbuf));
+#endif
     /* NOTREACHED */
 }
+#endif /* SCM_USE_STRING */
 
 static ScmObj
 read_symbol(ScmObj port)
@@ -889,7 +966,8 @@ read_symbol(ScmObj port)
         if (err != TOKEN_BUF_EXCEEDED)
             break;
         offset += tail_len;
-        LBUF_EXTEND(lbuf, SCM_LBUF_F_SYMBOL, LBUF_SIZE(lbuf) + MB_MAX_SIZE);
+        LBUF_EXTEND(lbuf, SCM_LBUF_F_SYMBOL,
+                    LBUF_SIZE(lbuf) + SCM_MB_CHAR_BUF_SIZE);
     }
 
     sym = scm_intern(LBUF_BUF(lbuf));
@@ -899,7 +977,7 @@ read_symbol(ScmObj port)
 }
 
 static ScmObj
-read_number_or_symbol(ScmObj port)
+read_number_or_peculiar(ScmObj port)
 {
     scm_ichar_t c;
     int err;
@@ -913,6 +991,7 @@ read_number_or_symbol(ScmObj port)
     SCM_ASSERT(ICHAR_ASCII_CLASS(c)
                & (SCM_CH_DIGIT | SCM_CH_PECULIAR_IDENTIFIER_CAND));
 
+#if SCM_USE_NUMBER
     if (ICHAR_NUMERICP(c))
         return read_number(port, 'd');
 
@@ -921,24 +1000,18 @@ read_number_or_symbol(ScmObj port)
         if (err == TOKEN_BUF_EXCEEDED)
             ERR("invalid number literal");
 
-
-        if (!buf[1]                           /* '+' or '-' */
-#if !SCM_STRICT_R5RS
-            /* FIXME: Obsolete with SRFI-75 style '|-sym| */
-            || (c == '-' && ICHAR_ALPHABETICP(buf[1]))  /* '-sym' */
-#endif
-            )
-        {
+        /* '+' or '-' */
+        if (!buf[1])
             return scm_intern(buf);
-        }
 
         return parse_number(port, buf, sizeof(buf), 'd');
     }
+#endif /* SCM_USE_NUMBER */
 
     if (c == '.') {
         read_token(port, &err, buf, sizeof(buf), SCM_CH_DELIMITER);
         if (strcmp(buf, "...") == 0)
-            return scm_intern(buf);
+            return SYM_ELLIPSIS;
         /* TODO: support numeric expressions when the numeric tower is
            implemented */
         ERR("invalid identifier: ~S", buf);
@@ -950,6 +1023,7 @@ read_number_or_symbol(ScmObj port)
     return read_symbol(port);
 }
 
+#if SCM_USE_NUMBER
 /* reads 'b123' part of #b123 */
 static ScmObj
 parse_number(ScmObj port, char *buf, size_t buf_size, char prefix)
@@ -990,9 +1064,10 @@ read_number(ScmObj port, char prefix)
 
     return parse_number(port, buf, sizeof(buf), prefix);
 }
+#endif /* SCM_USE_NUMBER */
 
 static ScmObj
-read_quote(ScmObj port, ScmObj quoter)
+read_quoted(ScmObj port, ScmObj quoter)
 {
     return SCM_LIST_2(quoter, read_sexpression(port));
 }
