@@ -2,7 +2,10 @@
  *  Filename : write.c
  *  About    : Object writer
  *
- *  Copyright (C) 2005-2006 Kazuki Ohta <mover AT hct.zaq.ne.jp>
+ *  Copyright (C) 2005      Kazuki Ohta <mover AT hct.zaq.ne.jp>
+ *  Copyright (C) 2005-2006 Jun Inoue <jun.lambda AT gmail.com>
+ *  Copyright (C) 2005-2006 YAMAMOTO Kengo <yamaken AT bp.iij4u.or.jp>
+ *  Copyright (c) 2007 SigScheme Project <uim AT freedesktop.org>
  *
  *  All rights reserved.
  *
@@ -41,7 +44,6 @@
 #include <stdarg.h>
 #include <string.h>
 
-#include "sigscheme-stdint.h"
 #include "sigscheme.h"
 #include "sigschemeinternal.h"
 
@@ -49,6 +51,14 @@
   File Local Macro Definitions
 =======================================*/
 #if SCM_USE_SRFI38
+#if !SCM_USE_STRING
+#define STRINGP(o)        0
+#define SCM_STRING_LEN(o) 0
+#endif
+#if !SCM_USE_VECTOR
+#undef  VECTORP
+#define VECTORP(o) 0
+#endif
 #define INTERESTINGP(obj)                                                    \
     (CONSP(obj)                                                              \
      || (STRINGP(obj) && SCM_STRING_LEN(obj))                                \
@@ -58,9 +68,10 @@
      || ERROBJP(obj))
 #define OCCUPIED(ent)      (!EQ((ent)->key, SCM_INVALID))
 #define HASH_EMPTY(table)  (!(table).used)
+/* datum index */
 #define DEFINING_DATUM     (-1)
 #define NONDEFINING_DATUM  0
-#define GET_DEFINDEX(x)    ((unsigned)(x) >> 1)
+/* flags */
 #define HASH_INSERT    1 /* insert key if it's not registered yet */
 #define HASH_FIND      0
 #endif /* SCM_USE_SRFI38 */
@@ -69,16 +80,17 @@
   File Local Type Definitions
 =======================================*/
 enum OutputType {
-    AS_WRITE,   /* string is enclosed by ", char is written using #\ notation */
-    AS_DISPLAY, /* string and char is written as-is */
-    UNKNOWN
+    UNKNOWN,
+    AS_WRITE,  /* string is enclosed by ", char is written using #\ notation */
+    AS_DISPLAY /* string and char is written as-is */
 };
 
 #if SCM_USE_SRFI38
 typedef size_t hashval_t;
+
 typedef struct {
     ScmObj key;
-    int datum;
+    scm_intobj_t datum;
 } hash_entry;
 
 typedef struct {
@@ -89,13 +101,15 @@ typedef struct {
 
 typedef struct {
     hash_table seen; /* a table of seen objects */
-    int next_index;  /* the next index to use for #N# */
+    scm_intobj_t next_index;  /* the next index to use for #N# */
 } write_ss_context;
 #endif /* SCM_USE_SRFI38 */
 
 /*=======================================
   Variable Definitions
 =======================================*/
+#include "functable-r5rs-write.c"
+
 SCM_DEFINE_EXPORTED_VARS(write);
 
 #if SCM_USE_SRFI38
@@ -112,15 +126,21 @@ SCM_DEFINE_STATIC_VARS(static_write);
 /*=======================================
   File Local Function Declarations
 =======================================*/
-static void write_internal(ScmObj port, ScmObj obj, enum OutputType otype);
-static void write_obj(ScmObj port, ScmObj obj, enum OutputType otype);
-static void write_char(ScmObj port, ScmObj obj, enum OutputType otype);
-static void write_string(ScmObj port, ScmObj obj, enum OutputType otype);
-static void write_list(ScmObj port, ScmObj lst, enum OutputType otype);
-static void write_vector(ScmObj port, ScmObj vec, enum OutputType otype);
-static void write_port(ScmObj port, ScmObj obj, enum OutputType otype);
-static void write_constant(ScmObj port, ScmObj obj, enum  OutputType otype);
-static void write_errobj(ScmObj port, ScmObj obj, enum  OutputType otype);
+static void write_internal (ScmObj port, ScmObj obj, enum OutputType otype);
+static void write_obj      (ScmObj port, ScmObj obj, enum OutputType otype);
+#if SCM_USE_CHAR
+static void write_char     (ScmObj port, ScmObj obj, enum OutputType otype);
+#endif
+#if SCM_USE_STRING
+static void write_string   (ScmObj port, ScmObj obj, enum OutputType otype);
+#endif
+static void write_list     (ScmObj port, ScmObj lst, enum OutputType otype);
+#if SCM_USE_VECTOR
+static void write_vector   (ScmObj port, ScmObj vec, enum OutputType otype);
+#endif
+static void write_port     (ScmObj port, ScmObj obj, enum OutputType otype);
+static void write_constant (ScmObj port, ScmObj obj, enum OutputType otype);
+static void write_errobj   (ScmObj port, ScmObj obj, enum OutputType otype);
 
 #if SCM_USE_HYGIENIC_MACRO
 static void write_farsymbol(ScmObj port, ScmObj obj, enum OutputType otype);
@@ -128,9 +148,10 @@ static void write_farsymbol(ScmObj port, ScmObj obj, enum OutputType otype);
 
 #if SCM_USE_SRFI38
 static void hash_grow(hash_table *tab);
-static hash_entry *hash_lookup(hash_table *tab, ScmObj key, int datum, int flag);
+static hash_entry *hash_lookup(hash_table *tab,
+                               ScmObj key, scm_intobj_t datum, int flag);
 static void write_ss_scan(ScmObj obj, write_ss_context *ctx);
-static int  get_shared_index(ScmObj obj);
+static scm_intobj_t get_shared_index(ScmObj obj);
 static void write_ss_internal(ScmObj port, ScmObj obj, enum OutputType otype);
 #endif /* SCM_USE_SRFI38 */
 
@@ -145,6 +166,8 @@ scm_init_writer(void)
     SCM_GLOBAL_VARS_INIT(static_write);
 #endif
 
+    scm_register_funcs(scm_functable_r5rs_write);
+
     /* To allow re-initialization of the interpreter, this variable must be
      * re-initialized by assignment. Initialized .data section does not work
      * for such situation.  -- YamaKen 2006-03-31 */
@@ -171,7 +194,7 @@ write_internal(ScmObj port, ScmObj obj, enum OutputType otype)
     ENSURE_PORT(port);
     SCM_ENSURE_LIVE_PORT(port);
     if (!(SCM_PORT_FLAG(port) & SCM_PORTFLAG_OUTPUT))
-        ERR("output port is required");
+        ERR_OBJ("output port required but got", port);
 
     write_obj(port, obj, otype);
     scm_port_flush(port);
@@ -184,23 +207,25 @@ write_obj(ScmObj port, ScmObj obj, enum OutputType otype)
 
 #if SCM_USE_SRFI38
     if (INTERESTINGP(obj)) {
-        int index = get_shared_index(obj);
+        scm_intobj_t index = get_shared_index(obj);
         if (index > 0) {
             /* defined datum */
-            scm_format(port, SCM_FMT_RAW_C, "#~D#", index);
+            scm_format(port, SCM_FMT_RAW_C, "#~ZU#", (size_t)index);
             return;
         }
         if (index < 0) {
             /* defining datum, with the new index negated */
-            scm_format(port, SCM_FMT_RAW_C, "#~D=", -index);
+            scm_format(port, SCM_FMT_RAW_C, "#~ZU=", (size_t)-index);
             /* Print it; the next time it'll be defined. */
         }
     }
 #endif
     switch (SCM_TYPE(obj)) {
+#if SCM_USE_INT
     case ScmInt:
         scm_format(port, SCM_FMT_RAW_C, "~MD", SCM_INT_VALUE(obj));
         break;
+#endif
     case ScmCons:
         if (ERROBJP(obj))
             write_errobj(port, obj, otype);
@@ -210,12 +235,16 @@ write_obj(ScmObj port, ScmObj obj, enum OutputType otype)
     case ScmSymbol:
         scm_port_puts(port, SCM_SYMBOL_NAME(obj));
         break;
+#if SCM_USE_CHAR
     case ScmChar:
         write_char(port, obj, otype);
         break;
+#endif
+#if SCM_USE_STRING
     case ScmString:
         write_string(port, obj, otype);
         break;
+#endif
     case ScmFunc:
         scm_port_puts(port, (SCM_SYNTAXP(obj)) ? "#<syntax " : "#<subr ");
         sym = scm_symbol_bound_to(obj);
@@ -261,21 +290,22 @@ write_obj(ScmObj port, ScmObj obj, enum OutputType otype)
         write_obj(port, SCM_CLOSURE_EXP(obj), otype);
         scm_port_put_char(port, '>');
         break;
+#if SCM_USE_VECTOR
     case ScmVector:
         write_vector(port, obj, otype);
         break;
+#endif
     case ScmPort:
         write_port(port, obj, otype);
         break;
+#if SCM_USE_CONTINUATION
     case ScmContinuation:
         scm_format(port, SCM_FMT_RAW_C, "#<continuation ~P>", (void *)obj);
         break;
+#endif
     case ScmValuePacket:
         scm_port_puts(port, "#<values ");
-        if (NULLP(SCM_VALUEPACKET_VALUES(obj)))
-            scm_port_puts(port, "()");
-        else
-            write_list(port, SCM_VALUEPACKET_VALUES(obj), otype);
+        write_obj(port, SCM_VALUEPACKET_VALUES(obj), otype);
 #if SCM_USE_VALUECONS
 #if SCM_USE_STORAGE_FATTY
         /* SCM_VALUEPACKET_VALUES() changes the type destructively */
@@ -289,6 +319,7 @@ write_obj(ScmObj port, ScmObj obj, enum OutputType otype)
     case ScmConstant:
         write_constant(port, obj, otype);
         break;
+#if SCM_USE_SSCM_EXTENSIONS
     case ScmCPointer:
         scm_format(port, SCM_FMT_RAW_C,
                    "#<c_pointer ~P>", SCM_C_POINTER_VALUE(obj));
@@ -298,11 +329,17 @@ write_obj(ScmObj port, ScmObj obj, enum OutputType otype)
                    "#<c_func_pointer ~P>",
                    (void *)(uintptr_t)SCM_C_FUNCPOINTER_VALUE(obj));
         break;
+#endif
+
+    case ScmRational:
+    case ScmReal:
+    case ScmComplex:
     default:
-        SCM_ASSERT(scm_false);
+        SCM_NOTREACHED;
     }
 }
 
+#if SCM_USE_CHAR
 static void
 write_char(ScmObj port, ScmObj obj, enum OutputType otype)
 {
@@ -332,10 +369,12 @@ write_char(ScmObj port, ScmObj obj, enum OutputType otype)
         break;
 
     default:
-        SCM_ASSERT(scm_false);
+        SCM_NOTREACHED;
     }
 }
+#endif /* SCM_USE_CHAR */
 
+#if SCM_USE_STRING
 static void
 write_string(ScmObj port, ScmObj obj, enum OutputType otype)
 {
@@ -384,7 +423,9 @@ write_string(ScmObj port, ScmObj obj, enum OutputType otype)
             continue2:
                 ;
             }
+#if SCM_USE_MULTIBYTE_CHAR
         }
+#endif
         scm_port_put_char(port, '\"'); /* closing doublequote */
         break;
 
@@ -393,9 +434,10 @@ write_string(ScmObj port, ScmObj obj, enum OutputType otype)
         break;
 
     default:
-        SCM_ASSERT(scm_false);
+        SCM_NOTREACHED;
     }
 }
+#endif /* SCM_USE_STRING */
 
 static void
 write_list(ScmObj port, ScmObj lst, enum OutputType otype)
@@ -403,7 +445,7 @@ write_list(ScmObj port, ScmObj lst, enum OutputType otype)
     ScmObj car;
 #if SCM_USE_SRFI38
     size_t necessary_close_parens;
-    int index;
+    scm_intobj_t index;
 #endif
     DECLARE_INTERNAL_FUNCTION("write");
 
@@ -412,10 +454,7 @@ write_list(ScmObj port, ScmObj lst, enum OutputType otype)
   cheap_recursion:
 #endif
 
-    if (NULLP(lst)) {
-        scm_port_puts(port, "()");
-        return;
-    }
+    SCM_ASSERT(CONSP(lst));
 
     scm_port_put_char(port, '(');
 
@@ -432,12 +471,12 @@ write_list(ScmObj port, ScmObj lst, enum OutputType otype)
         index = get_shared_index(lst);
         if (index > 0) {
             /* defined datum */
-            scm_format(port, SCM_FMT_RAW_C, ". #~D#", index);
+            scm_format(port, SCM_FMT_RAW_C, ". #~ZU#", (size_t)index);
             goto close_parens_and_return;
         }
         if (index < 0) {
             /* defining datum, with the new index negated */
-            scm_format(port, SCM_FMT_RAW_C, ". #~D=", -index);
+            scm_format(port, SCM_FMT_RAW_C, ". #~ZU=", (size_t)-index);
             necessary_close_parens++;
             goto cheap_recursion;
         }
@@ -458,6 +497,7 @@ write_list(ScmObj port, ScmObj lst, enum OutputType otype)
         scm_port_put_char(port, ')');
 }
 
+#if SCM_USE_VECTOR
 static void
 write_vector(ScmObj port, ScmObj vec, enum OutputType otype)
 {
@@ -476,6 +516,7 @@ write_vector(ScmObj port, ScmObj vec, enum OutputType otype)
 
     scm_port_put_char(port, ')');
 }
+#endif /* SCM_USE_VECTOR */
 
 static void
 write_port(ScmObj port, ScmObj obj, enum OutputType otype)
@@ -485,12 +526,11 @@ write_port(ScmObj port, ScmObj obj, enum OutputType otype)
     scm_port_puts(port, "#<");
 
     /* input or output */
-    /* print "i", "o" or "io" if bidirectional port */
+    /* print "iport", "oport" or "ioport" if bidirectional port */
     if (SCM_PORT_FLAG(obj) & SCM_PORTFLAG_INPUT)
         scm_port_put_char(port, 'i');
     if (SCM_PORT_FLAG(obj) & SCM_PORTFLAG_OUTPUT)
         scm_port_put_char(port, 'o');
-
     scm_port_puts(port, "port");
 
     /* file or string */
@@ -516,17 +556,13 @@ write_constant(ScmObj port, ScmObj obj, enum  OutputType otype)
     else if (EQ(obj, SCM_FALSE))
         str = "#f";
     else if (EQ(obj, SCM_EOF))
-#if SCM_COMPAT_SIOD_BUGS
-        str = "(eof)";
-#else
         str = "#<eof>";
-#endif
     else if (EQ(obj, SCM_UNBOUND))
         str = "#<unbound>";
     else if (EQ(obj, SCM_UNDEF))
         str = "#<undef>";
     else
-        SCM_ASSERT(scm_false);
+        SCM_NOTREACHED;
 
     scm_port_puts(port, str);
 }
@@ -551,12 +587,10 @@ write_errobj(ScmObj port, ScmObj obj, enum  OutputType otype)
 
     case AS_DISPLAY:
         scm_display(port, reason);
-        if (CONSP(objs))
-            scm_port_put_char(port, ':');
         break;
 
     default:
-        SCM_ASSERT(scm_false);
+        SCM_NOTREACHED;
     }
 
     FOR_EACH (elm, objs) {
@@ -606,7 +640,7 @@ hash_grow(hash_table *tab)
  * @return A pointer to the entry, or NULL if not found.
  */
 static hash_entry *
-hash_lookup(hash_table *tab, ScmObj key, int datum, int flag)
+hash_lookup(hash_table *tab, ScmObj key, scm_intobj_t datum, int flag)
 {
     size_t i;
     hashval_t hashval;
@@ -648,7 +682,7 @@ hash_lookup(hash_table *tab, ScmObj key, int datum, int flag)
     }
 
     /* A linear probe should always find a slot. */
-    SCM_ASSERT(scm_false);
+    SCM_NOTREACHED;
 }
 
 /**
@@ -659,7 +693,9 @@ hash_lookup(hash_table *tab, ScmObj key, int datum, int flag)
 static void
 write_ss_scan(ScmObj obj, write_ss_context *ctx)
 {
+#if SCM_USE_VECTOR
     scm_int_t i, len;
+#endif
     hash_entry *ent;
     ScmObj err_obj_tag, reason, objs, trace_stack;
     DECLARE_INTERNAL_FUNCTION("write-with-shared-structure");
@@ -702,6 +738,7 @@ write_ss_scan(ScmObj obj, write_ss_context *ctx)
 #if SCM_USE_VALUECONS
 #if SCM_USE_STORAGE_FATTY
             if (!SCM_NULLVALUESP(obj)) {
+                /* EQ(obj, SCM_VALUEPACKET_VALUES(obj)) */
                 write_ss_scan(CDR(SCM_VALUEPACKET_VALUES(obj)), ctx);
                 /* SCM_VALUEPACKET_VALUES() changes the type destructively */
                 SCM_ENTYPE(obj, ScmValuePacket);
@@ -714,10 +751,12 @@ write_ss_scan(ScmObj obj, write_ss_context *ctx)
 #endif /* SCM_USE_VALUECONS */
             break;
 
+#if SCM_USE_VECTOR
         case ScmVector:
             for (i = 0, len = SCM_VECTOR_LEN(obj); i < len; i++)
                 write_ss_scan(SCM_VECTOR_VEC(obj)[i], ctx);
             break;
+#endif /* SCM_USE_VECTOR */
 
         default:
             break;
@@ -731,7 +770,7 @@ write_ss_scan(ScmObj obj, write_ss_context *ctx)
  *         *additive inverse* of the index.  If obj is nondefining,
  *         return zero.
  */
-static int
+static scm_intobj_t
 get_shared_index(ScmObj obj)
 {
     hash_entry *ent;
@@ -742,7 +781,7 @@ get_shared_index(ScmObj obj)
         if (ent) {
             if (ent->datum == DEFINING_DATUM) {
                 ent->datum = l_write_ss_ctx->next_index++;
-                return - (ent->datum);
+                return -(ent->datum);
             }
             return ent->datum;
         }
@@ -754,7 +793,7 @@ static void
 write_ss_internal(ScmObj port, ScmObj obj, enum OutputType otype)
 {
     write_ss_context ctx = {{0}};
-    unsigned int i;
+    size_t i;
 
     ctx.next_index = 1;
     ctx.seen.size = 1 << 8; /* arbitrary initial size */
